import smtplib
from dataclasses import dataclass
from email.message import EmailMessage
from typing import Optional

from simple_mailer import exceptions


class NotConnectedError(Exception):
    """Not connected to the intended mail server"""


@dataclass
class Mailer:
    """An simple email client

    Attributes:
        host: the hostname of the server
        port: the port of the server
        use_tls: securely connect using TLS
    """

    host: str = "localhost"
    port: int = 465
    use_tls: bool = True
    _conn: smtplib.SMTP = Optional[None]

    def connect(self) -> "Mailer":
        """Connect to the remote server"""
        if self.host and self.port > 0:
            try:
                self._conn = smtplib.SMTP(host=self.host, port=self.port)
            except ConnectionRefusedError as exc:
                raise exceptions.MailServerError(
                    f"Could not connect to SMTP server "
                    f"at {self.host}:{self.port}: {exc.strerror}"
                )
            else:
                if self.use_tls:
                    self._conn.starttls()
                return self
        else:
            raise NotConnectedError(
                f"Cannot connect to: {self.host}:{self.port}"
            )

    def login(self, userid: str, passwd: str) -> "Mailer":
        """Login to the remote server

        Parameters:
            userid: the user id of the client
            passwd: the password of the client
        """
        if self._conn is None:
            raise NotConnectedError("Not connected. Please connect first.")
        else:
            self._conn.login(userid, passwd)
            return self

    def send_message(self, from_="", to="", subject="", body="") -> "Mailer":
        """Send an email message

        Parameters:
            from_: the sender email address
            to: the recipient email address
            subject: the subject of the email
            body: the body of the message
        """
        msg = EmailMessage()
        msg["Subject"] = subject
        msg["From"] = from_
        msg["To"] = to
        msg.set_content(body)
        self._conn.send_message(msg)
        return self

    def disconnect(self) -> "Mailer":
        """Disconnect from the remote server"""
        self._conn.close()
        return self



import json

import pytest
import webtest
from simple_mailer.config import settings
from simple_mailer.http import Location
from simple_mailer.tests.fixtures.smtpd import SMTPServerFixture
from simple_mailer.tests.helpers import with_environ_var


def test_root_path_has_link_to_mailer(test_http_client):
    response = test_http_client.get("/")
    assert response.status_code == 200
    data = json.loads(response.body.decode("utf8"))
    assert data["mailer"] == settings.MAILER_PATH


def test_post_urlencode_mail_app(smtpd: SMTPServerFixture, test_http_client):
    response = test_http_client.post(
        "/mail", {"email": "me@example.com", "subscribe_me": True}
    )
    assert response.status_code == 200
    assert smtpd.sent_mail


def test_post_json_mail_app(smtpd: SMTPServerFixture, test_http_client):
    response = test_http_client.post_json(
        "/mail", {"email": "me@example.com", "subscribe_me": True}
    )
    assert response.status_code == 200
    assert smtpd.sent_mail
    body = smtpd.sent_mail[0].body.decode("utf8")
    assert "timestamp_utc" in body


@with_environ_var("REDIRECT_URL", "{{ ORIGIN }}")
def test_post_with_referer(smtpd: SMTPServerFixture, test_http_client):
    response = test_http_client.post_json(
        "/mail", {"email": "me@example.com", "subscribe_me": True},
        headers={'ORIGIN': 'http://www.example.org/thank-you'}
    )
    assert response.status_code == 302
    assert response.headers["Location"] == "http://www.example.org/thank-you"


@with_environ_var("REDIRECT_URL", "{{ REFERER }}")
def test_post_with_referer(smtpd: SMTPServerFixture, test_http_client):
    response = test_http_client.post_json(
        "/mail", {"email": "me@example.com", "subscribe_me": True},
        headers={'REFERER': 'http://www.example.org/thank-you'}
    )
    assert response.status_code == 302
    assert response.headers["Location"] == "http://www.example.org/thank-you"


def test_post_empty_payload_is_denied(test_http_client):
    try:
        test_http_client.post_json("/mail", {})
    except webtest.app.AppError as err:
        assert "400 Bad Request" in err.args[0]


def test_location_url():
    loc = Location("www.example.com", "/bar/foo")
    assert loc.https_url == "https://www.example.com/bar/foo"


@with_environ_var("REDIRECT_URL", "https://www.example.org/target")
def test_if_redirect_url_set_response_is_302(test_http_client):
    response = test_http_client.post_json(
        "/mail", {"email": "me@example.com", "subscribe_me": True}
    )
    assert response.status_code == 302
    assert response.headers["Location"] == "https://www.example.org/target"


def test_debug_not_available_without_env_var(test_http_client):
    with pytest.raises(webtest.app.AppError):
        test_http_client.get("/debug")


@with_environ_var("ENABLE_DEBUG", "true")
def test_debug_available_with_env_var(test_http_client):
    root_data = json.loads(test_http_client.get("/").body)
    debug_path = root_data["debug"]
    assert settings.DEBUG_PATH == debug_path
    response = test_http_client.get(debug_path)
    assert response.status_code == 200
    debug_data = json.loads(response.body)
    assert "version" in debug_data
    assert "environment_variables" in debug_data
    vars = debug_data["environment_variables"]
    assert len(vars) == len(type(settings).get_defaults())


@with_environ_var("ENABLE_DEBUG", "true")
def test_debug_can_test_smtpd_connection(test_http_client):
    response = test_http_client.get(settings.DEBUG_PATH)
    debug_data = json.loads(response.body)
    assert "smtp_connection" in debug_data
    assert debug_data["smtp_connection"] == {"status": "OK"}



from simple_mailer.dispatcher import Dispatcher
from simple_mailer.tests.helpers import with_environ_var


def test_dispatcher_sends_email(smtpd, urlencoded_post_request):
    Dispatcher(data={"email": "me@example.com"}).dispatch()
    assert smtpd.sent_mail


def test_dispatcher_sends_text_templated_email(smtpd, urlencoded_post_request):
    Dispatcher(data={"email": "you@example.net", "msg": "hello!"}).dispatch()
    assert len(smtpd.sent_mail) == 1
    body = smtpd.sent_mail[0].body.decode("utf8")
    assert "Form submission" in body
    assert "msg: hello!" in body
    assert "timestamp_utc" in body


@with_environ_var("FIELDS_EXCLUDED", "subscribe_me")
def test_dispatcher_filter_unwanted_fields(smtpd, urlencoded_post_request):
    Dispatcher().parse_request(urlencoded_post_request).dispatch()
    assert len(smtpd.sent_mail) == 1
    body = smtpd.sent_mail[0].body.decode("utf8")
    assert "subscribe_me" not in body


@with_environ_var("FIELDS_INCLUDED", "subscribe_me")
def test_dispatcher_filter_wanted_fields(smtpd, urlencoded_post_request):
    Dispatcher().parse_request(urlencoded_post_request).dispatch()
    assert len(smtpd.sent_mail) == 1
    body = smtpd.sent_mail[0].body.decode("utf8")
    assert "email" not in body


@with_environ_var(
    "MAIL_SUBJECT", "IP {{metadata.client_ip}} sent a message: {{data.msg}}"
)
def test_mail_subject_can_be_defined_using_template_tags(smtpd):
    """The subject of the test mail can be configured using an env var"""
    dis = Dispatcher(data={"msg": "hello"}, metadata={"client_ip": "4.4.4.4"})
    dis.dispatch()
    body = smtpd.sent_mail[0].body.decode("utf8")
    assert "IP 4.4.4.4 sent a message: hello" in body



import datetime
import json
from urllib.parse import parse_qs

import bottle
from jinja2 import Template, UndefinedError, TemplateError
from simple_mailer import exceptions
from simple_mailer.captcha import CaptchaClient
from simple_mailer.config import settings
from simple_mailer.mailer import Mailer
from simple_mailer.utils import get_logger

log = get_logger(__name__)


class Dispatcher:
    """A controller that processes incoming HTTP requests and sends mail"""

    data: dict
    metadata: dict
    captcha_client: CaptchaClient

    def __init__(self, data=None, metadata=None):
        self.data = {} if data is None else data
        self.metadata = {} if metadata is None else metadata
        self.captcha_client = CaptchaClient.from_environment()

    def process_data(self) -> "Dispatcher":
        """Process the data, applying filters and manipulations appropriately
        """
        data = self.data

        log.debug(f"Received payload: {data}")

        fields_to_include = set(settings.FIELDS_INCLUDED)
        if fields_to_include:
            if self.captcha_client.key:
                fields_to_include.add(self.captcha_client.key)
            data = {k: v for k, v in data.items() if k in fields_to_include}
        fields_to_exclude = settings.FIELDS_EXCLUDED
        if fields_to_exclude:
            data = {
                k: v for k, v in data.items() if k not in fields_to_exclude
            }

        log.debug(f"After filters were applied, the payload was: {data}")

        if not data:
            raise exceptions.SubmittedDataInvalid("Need at least one field")
        self.data = data
        return self

    def parse_request(self, request: bottle.Request) -> "Dispatcher":
        """Extract and store the payload of a given HTTP request"""
        env = request.environ
        content_type = env["CONTENT_TYPE"]
        client_ip = env.get("HTTP_X_FORWARDED_FOR", env.get("REMOTE_ADDR", ""))
        log.info(
            f"Processing HTTP request from client with IP {client_ip} ..."
        )
        self.metadata = {
            "mailer_url": request.url,
            "origin": request.remote_addr or "",
            "client_ip": client_ip,
        }
        body = request.body.read().decode(request.POST.input_encoding)
        if content_type == "application/x-www-form-urlencoded":
            data = {k: (", ").join(v) for k, v in parse_qs(body).items()}
        elif content_type == "application/json":
            try:
                data = json.loads(body)
            except json.JSONDecodeError:
                err = log.error("Received invalid JSON")
                log.debug(f"Invalid JSON: {body}")
                raise exceptions.SubmittedDataInvalid(err)
            else:
                log.debug(f"Submitted payload: {data}")
        else:
            err = (
                f"Client sent request unsupported content type: {content_type}"
            )
            log.warning(err)
            raise exceptions.ContentTypeUnsupported(err)
        self.data = data
        self.process_data()
        return self

    def _get_templated_body(self) -> str:
        """Assemble and return the body of the message using the template"""
        tmpl_path = settings.MAIL_TEMPLATE_PATH
        if tmpl_path:
            try:
                with open(tmpl_path) as fd:
                    tmpl = Template(fd.read())
                    try:
                        return tmpl.render(
                            data=self.data, metadata=self.metadata
                        )
                    except UndefinedError as exc:
                        raise TemplateError(
                            f"The template did not define the required fields:"
                            f" {exc.message}"
                        )
            except IOError:
                err = (
                    f"Cannot open template file. "
                    f"Check if it exists and is readable."
                )
                log.error(err)
                raise exceptions.ConfigError(err)
        else:
            return json.dumps(self.data)

    def get_server(self) -> Mailer:
        """Get an instance of the mailer server"""
        return Mailer(
            host=settings.SMTP_HOST,
            port=settings.SMTP_PORT,
            use_tls=settings.USE_TLS,
        )

    def dispatch(self) -> None:
        """Dispatch a given HTTP request

        Returns true or false depending on the outcome.
        """
        self.captcha_client.validate_data(self.data)

        self.metadata.update(
            timestamp_utc=datetime.datetime.utcnow().isoformat()
        )

        cfg = {
            "from_": settings.FROM_ADDRESS,
            "to": settings.TO_ADDRESS,
            "subject": self.get_subject(),
            "body": self._get_templated_body(),
        }
        self.get_server().connect().send_message(**cfg).disconnect()
        log.info(
            f"Sent email to server {settings.SMTP_HOST}:{settings.SMTP_PORT}"
        )

    def get_subject(self) -> str:
        """Get the subject for the current email"""
        subject = settings.MAIL_SUBJECT
        if subject:
            tmpl = Template(subject)
            return tmpl.render(data=self.data, metadata=self.metadata)
        else:
            return subject



import logging
import os
from pathlib import Path
from typing import Optional, Any, Dict
from urllib.parse import urlparse

from bottle import request
from jinja2 import Template
from simple_mailer.exceptions import ConfigError
from simple_mailer.http import Location

_supported_log_level_names = ("DEBUG", "INFO", "WARN", "ERROR", "CRITICAL")


class BoolStr:
    """A factory class that can construct booleans from strings"""

    def __new__(cls, value, *args, **kwargs):
        if hasattr(value, "lower"):
            return bool(value.lower() == "true")
        else:
            return bool(value)


class TupleStr:
    sep = ","

    def __new__(cls, value, *args, **kwargs):
        if value and hasattr(value, "split"):
            return tuple(value.split(cls.sep))
        else:
            return tuple(value)


class _ConfigValueTypeChecker(type):
    """Check that the type definitions for the given class are supported"""

    supported_types = (int, float, str, bytes, bool, BoolStr, tuple, TupleStr)

    def __new__(cls, name, bases, ns):
        try:
            annotations: dict = ns["__annotations__"]
        except KeyError:
            pass
        else:
            for key, typ in annotations.items():
                if typ not in cls.supported_types:
                    raise TypeError(
                        f'Unsupported config variable type "{typ}" '
                        f'for key "{key}". '
                        f"Supported types are: {cls.supported_types}"
                    )

        return super().__new__(cls, name, bases, ns)


class _ConfigurationSettings:
    """A basic, lazily loaded configuration object."""

    class Defaults(metaclass=_ConfigValueTypeChecker):
        """Default values need to be specified for each configuration var.

        If not specified, the default type will be str.

        Only certain types are supported, e.g. str, float, bool, etc.
        See the metaclass for more information."""

        SMTP_HOST: str = "localhost"
        SMTP_PORT: int = 465
        SMTP_USERID: str = ""
        SMTP_PASSWD: str = ""
        TO_ADDRESS: str = ""
        FROM_ADDRESS: str = ""
        MAIL_SUBJECT: str = ""
        MAIL_TEMPLATE_PATH = (
                Path(__file__).parent / "templates" / "default.txt"
        )
        MAILER_PATH: str = "/mail"
        USE_TLS: BoolStr = BoolStr(True)
        CAPTCHA_TYPE: str = ""
        CAPTCHA_SECRET: str = ""
        CAPTCHA_VERIFY_URL: str = ""
        REDIRECT_URL: str = ""
        FIELDS_INCLUDED: TupleStr = TupleStr("")
        FIELDS_EXCLUDED: TupleStr = TupleStr("")
        ENABLE_DEBUG: BoolStr = BoolStr(False)
        DEBUG_PATH: str = "/debug"
        LOG_LEVEL: str = "WARN"

    def _get(self, name: str) -> Any:
        try:
            val = os.environ.get(name, getattr(self.Defaults, name))
        except AttributeError:
            raise ConfigError(f"Unsupported configuration parameter: {name}")
        else:
            typ = self.Defaults.__annotations__.get(name, str)
            try:
                return typ(val)
            except (TypeError, ValueError) as exc:
                raise ConfigError(
                    f'Configuration error: cannot convert value "{val}" '
                    f'for variable named "{name}" into type "{typ.__name__}": '
                    f"{exc}"
                )

    def __getattr__(self, name: str) -> Any:
        return self._get(name)

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    @property
    def CAPTCHA_VERIFY_LOCATION(self) -> Optional[Location]:
        """The location of the captcha verification site"""
        overridden_url = self.CAPTCHA_VERIFY_URL
        if overridden_url:
            parsed = urlparse(overridden_url)
            return Location(parsed.hostname, parsed.path)
        return None

    @property
    def LOG_LEVEL(self) -> int:
        name = self._get("LOG_LEVEL").upper()
        if name in _supported_log_level_names:
            return getattr(logging, name)
        else:
            raise ConfigError("Unsupported level name: " + name)

    @property
    def REDIRECT_URL(self) -> str:
        """The URL where the client will be redirected

        Can also be one of the following template tags: REFERER, ORIGIN
        """
        redirect_url = self._get("REDIRECT_URL")
        if "{{" in redirect_url:
            tmpl = Template(redirect_url)
            return tmpl.render(
                ORIGIN=request.headers.get('ORIGIN', '{{ ORIGIN }}'),
                REFERER=request.headers.get('REFERER', '{{ REFERER }}')
            )
        else:
            return redirect_url

    @classmethod
    def get_defaults(cls) -> Dict:
        """List the supported variable names"""
        ns = cls.Defaults.__dict__
        return {k: v for k, v in ns.items() if not k.startswith("__")}


settings = _ConfigurationSettings()
__all__ = (settings.__name__,)



import http.client
import json
from dataclasses import dataclass
from typing import Dict
from urllib.parse import urlencode

from simple_mailer import exceptions
from simple_mailer.config import settings
from simple_mailer.http import Location
from simple_mailer.utils import get_logger

log = get_logger(__name__)


@dataclass
class CaptchaClient:
    """An generic object that verifies a given captcha response
    """

    protocol_name: str = "noop"
    key: str = ""
    location = None

    def extract_response(self, data: Dict) -> str:
        """Extract the correct response from"""
        try:
            return data[self.key]
        except KeyError:
            err = (
                f"The expected response for "
                f"captcha protocol '{self.protocol_name}' was not found. "
                f"Expected a field named {self.key}."
            )
            log.error(err)
            raise exceptions.MissingCaptchaResponse(err)

    def validate_data(self, data: Dict) -> None:
        """Introspect the given data to infer and validate the response"""
        pass

    @staticmethod
    def from_environment() -> "CaptchaClient":
        """Choose and create a captcha client by introspecting the environment
        """
        protocol = settings.CAPTCHA_TYPE
        if not protocol:
            log.debug("No captcha protocol configured for use")
            return CaptchaClient()
        elif protocol == Recaptchav3Client.protocol_name:
            loc = settings.CAPTCHA_VERIFY_LOCATION
            if loc is None:
                client = Recaptchav3Client()
            else:
                client = Recaptchav3Client(location=loc)
            log.debug(
                f"Using captcha protocol {client.protocol_name} with "
                f"verification URL at {client.location.https_url}"
            )
            return client
        else:
            err = (
                f"Configuration Error: unsupported Captcha protocol: "
                f"{protocol}"
            )
            log.error(err)
            raise exceptions.UnknownCaptchaProtocol(err)


@dataclass
class Recaptchav3Client(CaptchaClient):
    """A Recaptchav3 client"""

    protocol_name: str = "recaptchav3"
    key: str = "g-recaptcha-response"
    location: Location = Location(
        "www.google.com", "/recaptcha/api/siteverify"
    )

    def validate_data(self, data: Dict) -> None:
        """Connect to the server and validate the given response"""
        resp = self.extract_response(data)
        params = {"secret": settings.CAPTCHA_SECRET, "response": resp}
        headers = {
            "Content-type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
        log.debug(f"Validating catpcha response data: {params}")
        conn = http.client.HTTPSConnection(self.location.hostname)
        log.debug(
            f"Sending captcha verification POST request to "
            f"{self.location.https_url} ..."
        )
        try:
            conn.request(
                "POST", self.location.https_url, urlencode(params), headers
            )
        except ConnectionRefusedError as exc:
            raise exceptions.CaptchaServerConnectionRefused(
                f'Could not connect to the Captcha server '
                f'at "{self.location.path}", reason: {exc}'
            )
        http_response = conn.getresponse()
        log.debug(f"Got {http_response.status} response from catpcha server.")
        if http_response.status == 200:
            data = json.load(http_response)
            if data["success"]:
                log.debug(f"Captcha response was validated successfully.")
                return None
            else:
                err = (
                    f"Captcha response failed validation"
                )
                log.debug(err)
                log.debug(f'Got this data from server: {data}')
                log.warning("Client failed captcha verification.")
                raise exceptions.InvalidCaptchaResponse(err)
        err = (
            f"The captcha response verification has failed. "
            f"The challenge response provided in the POST data was: {resp}"
        )
        log.debug(err)
        log.warning("Client failed captcha verification.")
        raise exceptions.FailedCaptchaResponse(err)



import json

from bottle import request, response, default_app, AppStack, get, post, run
from simple_mailer import checks
from simple_mailer import exceptions
from simple_mailer.config import settings
from simple_mailer.dispatcher import Dispatcher
from simple_mailer.utils import cloak, get_logger

log = get_logger(__name__)


@post(settings.MAILER_PATH)
def mail() -> str:
    """A resource that can send mail"""
    log.info(f"Got a new submission from client with IP {request.remote_addr}")
    try:
        Dispatcher().parse_request(request).dispatch()
    except exceptions.ContentTypeUnsupported as exc:
        response.status = 400
        return str(exc)
    except exceptions.SubmittedDataInvalid as exc:
        response.status = 400
        return str(exc)
    except exceptions.BaseSimpleMailerException as exc:
        response.status = 503
        return str(exc)
    else:
        redirect_url = settings.REDIRECT_URL
        if redirect_url:
            response.status = 302
            response.headers["Location"] = redirect_url
        else:
            response.status = 200
        return "OK"


@get(settings.DEBUG_PATH)
def debug() -> str:
    ns = {}
    if settings.ENABLE_DEBUG:
        ns.update(
            {
                "version": checks.get_version(),
                "environment_variables": cloak(checks.get_env_variables()),
                "smtp_connection": checks.get_smtp_connection(),
            }
        )
    else:
        response.status = 404
        ns["errors"] = "Debug not enabled"
    return json.dumps(ns)


@get("/")
def root() -> str:
    """The root resource"""
    ns = {"mailer": settings.MAILER_PATH}
    if settings.ENABLE_DEBUG:
        ns["debug"] = settings.DEBUG_PATH
    return json.dumps(ns)


def get_application() -> AppStack:
    """Get the default Bottle application"""
    return default_app()


def run_application() -> None:
    run(host="localhost", port=8080)


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--host", default="localhost", required=False, dest="host"
    )
    parser.add_argument(
        "--post", default=8080, type=int, required=False, dest="port"
    )
    parsed = parser.parse_args()
    run(host=parsed.host, port=parsed.port)


if __name__ == "__main__":
    main()
